package ircclient

import (
	"fmt"
	"net"
	"reflect"
	"strconv"
	"strings"
	"syscall"
	"testing"
)

func TestParse(t *testing.T) {
	var Msg IRCMsg
	var Expect map[string]IRCMsg = map[string]IRCMsg{
		"PING :78629A0F5F3F164F": IRCMsg{
			Cmd: "PING", Msg: "78629A0F5F3F164F", MsgParts: []string{"PING"}},
		":irc.red-green.com 001 test :Welcome to the RedGreen IRC Network": IRCMsg{
			From: "irc.red-green.com", Cmd: "001", To: "test",
			Msg:      "Welcome to the RedGreen IRC Network",
			MsgParts: []string{":irc.red-green.com", "001", "test"}},
		":irc.red-green.com 375 test :- irc.red-green.com Message of the Day -": IRCMsg{
			From: "irc.red-green.com", Cmd: "375", To: "test",
			Msg:      "- irc.red-green.com Message of the Day -",
			MsgParts: []string{":irc.red-green.com", "375", "test"}},
		":irc.red-green.com 433 :Nick already in use": IRCMsg{
			From: "irc.red-green.com", Cmd: "433",
			Msg:      "Nick already in use",
			MsgParts: []string{":irc.red-green.com", "433"}},
		":NickServ!services@services.red-green.com NOTICE test :Password accepted - you are now recognized.": IRCMsg{
			From: "NickServ", Cmd: "NOTICE", To: "test",
			Msg: "Password accepted - you are now recognized.",
			MsgParts: []string{":NickServ!services@services.red-green.com",
				"NOTICE", "test"}},
		":irc.red-green.com 005 meow-bot HCN INVEX KICKLEN=307 KNOCK MAP MAXCHANNELS=10 MAXLIST=b:60,e:60,I:60 MAXNICKLEN=30 MINNICKLEN=0 MODES=12 NAMESX NETWORK=RedGreen :are supported by this server": IRCMsg{
			From: "irc.red-green.com", Cmd: "005", To: "meow-bot",
			Msg: "are supported by this server",
			MsgParts: []string{":irc.red-green.com", "005", "meow-bot", "HCN", "INVEX",
				"KICKLEN=307", "KNOCK", "MAP", "MAXCHANNELS=10", "MAXLIST=b:60,e:60,I:60",
				"MAXNICKLEN=30", "MINNICKLEN=0", "MODES=12", "NAMESX", "NETWORK=RedGreen"}},
		":irc.red-green.com 353 meow-bot = #backstage :meow-bot performer5 performer4 performer3 performer1 performer2 Stage_Manager Apollo ~bugz": IRCMsg{
			From: "irc.red-green.com", Cmd: "353", To: "meow-bot",
			Msg:      "meow-bot performer5 performer4 performer3 performer1 performer2 Stage_Manager Apollo ~bugz",
			MsgParts: []string{":irc.red-green.com", "353", "meow-bot", "=", "#backstage"}},
	}

	for line, msgout := range Expect {
		Msg = IRCParse(line)
		var fields []string = []string{"To", "From", "Cmd", "Msg"}
		msg_value := reflect.ValueOf(Msg)
		msgout_value := reflect.ValueOf(msgout)

		for _, field := range fields {
			var got string
			var expect string
			got = msg_value.FieldByName(field).String()
			expect = msgout_value.FieldByName(field).String()
			if got != expect {
				t.Errorf("%s %s: got %s, expected %s", line, field, got, expect)
			}
		}

		/*
			if Msg.To != msgout.To {
				t.Errorf("%s To: got %s, expected %s", line, Msg.To, msgout.To)
			}
			if Msg.From != msgout.From {
				t.Errorf("%s From: got %s, expected %s", line, Msg.From, msgout.From)
			}
			if Msg.Cmd != msgout.Cmd {
				t.Errorf("%s Cmd: got %s, expected %s", line, Msg.Cmd, msgout.Cmd)
			}
			if Msg.Msg != msgout.Msg {
				t.Errorf("%s Msg: got %s, expected %s", line, Msg.Msg, msgout.Msg)
			}
		*/
		if len(Msg.MsgParts) != len(msgout.MsgParts) {
			t.Errorf("%s MsgParts got len %d, expected len %d", line, len(Msg.MsgParts), len(msgout.MsgParts))
			// Length didn't match, don't need to compare each part.
		} else {
			// Length matches, compare each part
			for idx := range Msg.MsgParts {
				if Msg.MsgParts[idx] != msgout.MsgParts[idx] {
					t.Errorf("%s MsgParts[%d] %s, expected %s", line, idx, Msg.MsgParts[idx], msgout.MsgParts[idx])
				}
			}
		}
	}
}

func TestConnect(t *testing.T) {
	var config IRCConfig = IRCConfig{Nick: "test",
		Username:       "test",
		Realname:       "testing",
		Password:       "12345",
		ServerPassword: "allow"}
	var listen net.Listener
	var address string

	listen, address = setupSocket()
	var parts []string = strings.Split(address, ":")

	config.Hostname = parts[0]
	config.Port, _ = strconv.Atoi(parts[1])
	go ircServer(listen, t, &config)
	var FromIRC chan IRCMsg

	FromIRC = make(chan IRCMsg)
	config.ReadChannel = FromIRC

	config.Connect()
	defer config.Close()

	var Msg IRCMsg
	var motd, identify, support bool
	var isupport map[string]string = map[string]string{
		"AWAYLEN": "307", "BOT": "B", "CASEMAPPING": "ascii",
		"CHANLIMIT": "#:10", "CHANMODES": "beI,kLf,lH,psmntirzMQNRTOVKDdGPZSCc",
		"CHANNELLEN": "32", "CHANTYPES": "#", "CLIENTTAGDENY": "*,-draft/typing,-typing",
		"DEAF": "d", "ELIST": "MNUCT", "EXCEPTS": "", "EXTBAN": "~,GptmTSOcarnqjf",
		"HCN": "", "INVEX": "", "KICKLEN": "307", "KNOCK": "", "MAP": "", "MAXCHANNELS": "10",
		"MAXLIST": "b:60,e:60,I:60", "MAXNICKLEN": "30", "MINNICKLEN": "0", "MODES": "12",
		"NAMESX": "", "NETWORK": "RedGreen", "NICKLEN": "30", "PREFIX": "(qaohv)~&@%+",
		"QUITLEN": "307", "SAFELIST": "", "SILENCE": "15", "STATUSMSG": "~&@%+",
		"TARGMAX":  "DCCALLOW:,ISON:,JOIN:,KICK:4,KILL:,LIST:,NAMES:1,NOTICE:1,PART:,PRIVMSG:4,SAJOIN:,SAPART:,TAGMSG:1,USERHOST:,USERIP:,WATCH:,WHOIS:1,WHOWAS:1",
		"TOPICLEN": "360", "UHNAMES": "", "USERIP": "", "WALLCHOPS": "", "WATCH": "128",
		"WATCHOPTS": "A", "WHOX": "",
	}

	for Msg = range FromIRC {
		if Msg.Cmd == "EndMOTD" {
			t.Log("Got EndMOTD")
			motd = true

			support = true
			// Verify we parsed 005 ISupport:
			for key, value := range isupport {
				var got string
				var has bool
				got, has = config.ISupport[key]
				if !has {
					t.Errorf("Missing ISUPPORT[%s] expected (%s)", key, value)
					support = false
				} else {
					if got != value {
						t.Errorf("ISUPPORT[%s] got %s, expected %s", key, got, value)
						support = false
					}
				}
			}

		}

		if Msg.Cmd == "Identified" {
			t.Log("Identified")
			identify = true
		}
	}

	if !motd {
		t.Error("Missing EndMOTD")
	}
	if !identify {
		t.Error("Missing Identified")
	}
	if !support {
		t.Error("Missing ISupport")
	}
	if config.MyNick != config.Nick {
		t.Errorf("Got %s, Expected %s", config.MyNick, config.Nick)
	}

}

func TestConnectNickInUse(t *testing.T) {
	var config IRCConfig = IRCConfig{Nick: "bad",
		Username:    "test",
		Realname:    "testing",
		Flood_Num:   1,
		Flood_Delay: 20,
	}
	var listen net.Listener
	var address string

	listen, address = setupSocket()
	var parts []string = strings.Split(address, ":")

	config.Hostname = parts[0]
	config.Port, _ = strconv.Atoi(parts[1])
	go ircServer(listen, t, &config)
	var FromIRC chan IRCMsg

	FromIRC = make(chan IRCMsg)
	config.ReadChannel = FromIRC

	config.Connect()
	defer config.Close()

	var Msg IRCMsg
	var motd, identify bool
	var missing int

	for Msg = range FromIRC {
		if Msg.Cmd == "EndMOTD" {
			t.Log("Got EndMOTD")
			motd = true
			config.WriteTo("missing", "PRIVMSG missing :Missing user")
			config.WriteTo("missing", "PRIVMSG missing :Missing user")
			config.WriteTo("missing", "PRIVMSG missing :Missing user")
			config.WriteTo("#missing", "PRIVMSG #missing :Missing channel")
		}
		if Msg.Cmd == "Identified" {
			t.Log("Identified")
			identify = true
		}
		if Msg.Cmd == "404" || Msg.Cmd == "401" {
			missing++
		}
	}

	if !motd {
		t.Error("Missing EndMOTD")
	}
	if identify {
		t.Error("Should not have been Identified")
	}
	if missing < 2 {
		t.Errorf("Missing should have been 2, was %d", missing)
	}
	if config.MyNick == config.Nick {
		t.Errorf("Nick should be different: Got %s, Didn't Expect %s", config.MyNick, config.Nick)
	}
}
func TestConnectTLS(t *testing.T) {
	var config IRCConfig = IRCConfig{Nick: "test",
		Username:       "test",
		Realname:       "testing",
		Password:       "12345",
		UseTLS:         true,
		UseSASL:        true,
		Insecure:       true,
		ServerPassword: "allow"}
	var listen net.Listener
	var address string

	listen, address = setupTLSSocket()
	var parts []string = strings.Split(address, ":")

	config.Hostname = parts[0]
	config.Port, _ = strconv.Atoi(parts[1])
	go ircServer(listen, t, &config)
	var FromIRC chan IRCMsg

	FromIRC = make(chan IRCMsg)
	config.ReadChannel = FromIRC

	config.Connect()
	defer config.Close()

	var Msg IRCMsg
	var motd, identify bool
	for Msg = range FromIRC {
		if Msg.Cmd == "EndMOTD" {
			t.Log("Got EndMOTD")
			motd = true
		}
		if Msg.Cmd == "Identified" {
			t.Log("Identified")
			identify = true
		}
	}

	if !motd {
		t.Error("Missing EndMOTD")
	}
	if !identify {
		t.Error("Missing Identified")
	}

	if config.MyNick != config.Nick {
		t.Errorf("Got %s, Expected %s", config.MyNick, config.Nick)
	}

}

func TestConnectAutojoin(t *testing.T) {
	var config IRCConfig = IRCConfig{Nick: "test",
		Username:    "test",
		Realname:    "testing",
		Password:    "12345",
		UseTLS:      true,
		UseSASL:     true,
		Insecure:    true,
		AutoJoin:    []string{"#chat", "#test"},
		Flood_Num:   2,
		Flood_Delay: 10,
		Version:     "Test",
	}
	var listen net.Listener
	var address string

	listen, address = setupTLSSocket()
	var parts []string = strings.Split(address, ":")

	config.Hostname = parts[0]
	config.Port, _ = strconv.Atoi(parts[1])
	go ircServer(listen, t, &config)
	var FromIRC chan IRCMsg

	FromIRC = make(chan IRCMsg)
	config.ReadChannel = FromIRC

	config.Connect()
	defer config.Close()

	var Msg IRCMsg
	var motd, identify bool
	var joins int
	var expect string
	var ctcpExpect []string = []string{"VERSION",
		"TIME",
		"PING 12345",
	}
	var noticeExpect []string = []string{"Testing",
		fmt.Sprintf("VERSION %s red-green.com/irc-client", config.Version),
		"TIME ",
		"PING 12345",
	}

	for Msg = range FromIRC {
		/*
			if (Msg.Cmd == "ACTION") || (Msg.Cmd == "NOTICE") {
				t.Log(Msg)
			}
		*/
		if Msg.Cmd == "EndMOTD" {
			t.Log("Got EndMOTD")
			motd = true
		}
		if Msg.Cmd == "Identified" {
			t.Log("Identified")
			identify = true
		}
		if Msg.Cmd == "JOIN" {
			joins++
			if joins == 2 {
				// messages set to echo are returned to us
				config.WriteTo("echo", "PRIVMSG echo :\x01VERSION\x01")
				config.WriteTo("echo", "PRIVMSG echo :\x01TIME\x01")
				config.WriteTo("echo", "PRIVMSG echo :\x01PING 12345\x01")
				config.Action("echo", "dances.")
				config.Notice("echo", "Testing")
				config.Msg("#test", "Message 1")
				config.Msg("#test", "Message 2")
				config.PriorityWrite("PRIVMSG #test :Message")
			}
		}
		if Msg.Cmd == "CTCP" {
			expect = ctcpExpect[0]
			ctcpExpect = ctcpExpect[1:]
			if Msg.Msg != expect {
				t.Errorf("CTCP Got %s, Expected %s", Msg.Msg, expect)
			}
		}
		if Msg.Cmd == "NOTICE" {
			expect = noticeExpect[0]
			if expect != "Testing" {
				expect = "\x01" + expect
			}
			noticeExpect = noticeExpect[1:]
			if !strings.HasPrefix(Msg.Msg, expect) {
				t.Errorf("NOTICE Got [%s], Expected [%s]", Msg.Msg, expect)
			}
		}
		if Msg.Cmd == "ACTION" {
			expect = "dances."
			if Msg.Msg != expect {
				t.Errorf("ACTION Got %s, Expected %s", Msg.Msg, expect)
			}
		}
	}

	if joins != 2 {
		t.Errorf("Expected to autojoin 2 channels, got %d", joins)
	}
	if !motd {
		t.Error("Missing EndMOTD")
	}
	if !identify {
		t.Error("Missing Identified")
	}
	if len(noticeExpect) != 0 {
		t.Errorf("Expected more NOTICEs (%d)", len(noticeExpect))
	}
	if len(ctcpExpect) != 0 {
		t.Errorf("Expected more CTCPs (%d)", len(ctcpExpect))
	}

	if config.MyNick != config.Nick {
		t.Errorf("Got %s, Expected %s", config.MyNick, config.Nick)
	}

}

func TestConnectKick(t *testing.T) {
	var config IRCConfig = IRCConfig{Nick: "test",
		Username:    "test",
		Realname:    "testing",
		Password:    "12345",
		UseTLS:      true,
		UseSASL:     true,
		Insecure:    true,
		AutoJoin:    []string{"#chat", "#test", "#kick"},
		RejoinDelay: 1,
		Flood_Num:   2,
		Flood_Delay: 10,
	}
	var listen net.Listener
	var address string

	listen, address = setupTLSSocket()
	var parts []string = strings.Split(address, ":")

	config.Hostname = parts[0]
	config.Port, _ = strconv.Atoi(parts[1])

	// Save and Restore abortAfter
	var abortPrev = abortAfter
	defer func() { abortAfter = abortPrev }()
	abortAfter = 500
	go ircServer(listen, t, &config)

	var FromIRC chan IRCMsg

	FromIRC = make(chan IRCMsg)
	config.ReadChannel = FromIRC

	config.Connect()
	defer config.Close()

	var Msg IRCMsg
	var motd, identify bool
	var joins int

	for Msg = range FromIRC {
		if Msg.Cmd == "EndMOTD" {
			t.Log("Got EndMOTD")
			motd = true
		}
		if Msg.Cmd == "Identified" {
			t.Log("Identified")
			identify = true
		}
		if Msg.Cmd == "JOIN" {
			joins++
			if joins == 4 {
				config.PriorityWrite("NICK something")
			}
		}
	}

	// 3 channels joined, 1 kick, +1 join=4.
	if joins != 4 {
		t.Errorf("Expected to join 4 times, got %d", joins)
	}
	if !motd {
		t.Error("Missing EndMOTD")
	}
	if !identify {
		t.Error("Missing Identified")
	}
	if config.MyNick != "something" {
		t.Errorf("Expected nick to be something, not %s", config.MyNick)
	}
}

func TestConnectSignal(t *testing.T) {
	var exit bool

	var onexit func() = func() { exit = true }

	var config IRCConfig = IRCConfig{Nick: "test",
		Username:    "test",
		Realname:    "testing",
		Password:    "12345",
		UseTLS:      true,
		UseSASL:     true,
		Insecure:    true,
		AutoJoin:    []string{"#chat"},
		Flood_Num:   2,
		Flood_Delay: 10,
		OnExit:      onexit,
	}
	var listen net.Listener
	var address string

	listen, address = setupTLSSocket()
	var parts []string = strings.Split(address, ":")

	config.Hostname = parts[0]
	config.Port, _ = strconv.Atoi(parts[1])
	go ircServer(listen, t, &config)
	var FromIRC chan IRCMsg

	FromIRC = make(chan IRCMsg)
	config.ReadChannel = FromIRC

	config.Connect()
	// defer config.Close()

	var Msg IRCMsg
	var motd, identify bool

	for Msg = range FromIRC {
		if Msg.Cmd == "EndMOTD" {
			t.Log("Got EndMOTD")
			motd = true
			// config.PriorityWrite("NICK something")
		}
		if Msg.Cmd == "Identified" {
			t.Log("Identified")
			identify = true
		}
		if Msg.Cmd == "JOIN" {
			// Ok, we've joined.  Test the signal
			e := syscall.Kill(syscall.Getpid(), syscall.SIGTERM)
			t.Log("Kill:", e)
		}
	}

	if !motd {
		t.Error("Missing EndMOTD")
	}
	if !identify {
		t.Error("Missing Identified")
	}

	config.Close()

	if !exit {
		t.Error("AtExit wasn't called.")
	}
}

func TestMatch(t *testing.T) {
	var testdata map[string]string = map[string]string{"TEST": "test",
		"TEst": "teST", "[bugz]": "{BUGZ}"}

	for nick1, nick2 := range testdata {
		if !Match(nick1, nick2) {
			t.Errorf("Match %s [%s] != %s [%s] Failed match!", nick1, NameLower(nick1), nick2, NameLower(nick2))
		}
	}
}
